(************************************************************************************
Filename    :   Win32_RoomTiny_Main.cpp
Content     :   First-person view test application for Oculus Rift
Created     :   18th Dec 2014
Authors     :   Tom Heath
Copyright   :   Copyright 2012 Oculus, Inc. All Rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
/// A simple sample to demonstrate the power and flexibility of the 'layers'
/// functionality now present in the SDK.
/// This isn't intended as a sample of best practise or good use case, as such.
/// Just showing how to set up layers with different characteristics.
/// Hold the '1' key to remove the smaller layer
/// Hold the '2' key to lower framerate on the outer layer - note as you translate,
/// there are issues because of the lower framerate on that outer layer.
/// Hold the '3' key is freeze the 1st layer
/// Hold the '4' key is freeze the 2nd layer
unit UDemoLayers;

interface

procedure Main;

implementation

uses
  System.Math, OVR_CAPI, UDirectXAppUtil, UBasicVR;

type
  TLayers = class(TBasicVR)
  private
    clock: Integer;
  public
    constructor Create;
    procedure MainLoop; override;
  end;

constructor TLayers.Create;
begin
  inherited Create('Layers');
end;

procedure TLayers.MainLoop;
var
  newFOV: array[0..1] of TOvrFovPort;
  updateLayer0, updateLayer1: Boolean;
  eye: NativeInt;
begin
  // Create a small FOV
  newFOV[0].DownTan := 0.5;
  newFOV[0].UpTan := 0.5;
  newFOV[0].LeftTan := 0.5;
  newFOV[0].RightTan := 0.5;
  newFOV[1] := newFOV[0];

  // Make layers, with 2nd one having smaller FOV, and the first having lower resolution
  Layer[0] := TVRLayer.Create(Session, nil, 0.33);
  Layer[1] := TVRLayer.Create(Session, @newFOV);

  // Main loop
  while HandleMessages do
  begin
    // Lets use a clock, and user input, to decide when each layer is updated
    Inc(clock);
    updateLayer0 := True;
    updateLayer1 := True;
    if DIRECTX.Key['2'] and ((clock mod 4) <> 0) then updateLayer0 := False;
    if DIRECTX.Key['3'] then updateLayer0 := False;
    if DIRECTX.Key['4'] then updateLayer1 := False;

    ActionFromInput;

    if updateLayer0 then Layer[0].GetEyePoses;
    if updateLayer1 then Layer[1].GetEyePoses;

    for eye := 0 to 1 do
    begin
      if updateLayer0 then Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye, nil, nil, 1, 1, 1, 0.8, 1);
      if updateLayer1 then Layer[1].RenderSceneToEyeBuffer(MainCam, RoomScene, eye, nil, nil, 1, 1, 1, 1, 1);
    end;

    Layer[0].PrepareLayerHeader;
    Layer[1].PrepareLayerHeader;

    // Press 1 to show just one layer
    if DIRECTX.Key['1'] then DistortAndPresent(1)
    else                     DistortAndPresent(2);
  end;
end;

procedure Main;
var
  app: TLayers;
begin
  app := TLayers.Create;
  try
    ExitCode := app.Run;
  finally
    app.Free;
  end;
end;

end.