(************************************************************************************
Filename    :   Win32_RoomTiny_Main.cpp
Content     :   First-person view test application for Oculus Rift
Created     :   18th Dec 2014
Authors     :   Tom Heath
Copyright   :   Copyright 2012 Oculus, Inc. All Rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
/// This sample demonstrates multi-sample anti-aliasing,
/// which can be compared to its absence by holding the '1' key.
unit UDemoMSAA;

interface

procedure Main;

implementation

uses
  Winapi.D3D11, Winapi.DxgiFormat, OVR_CAPI, OVR_CAPI_D3D, UDirectXAppUtil, UBasicVR;

type
  TMSAA = class(TBasicVR)
  public
    constructor Create;
    procedure MainLoop; override;
  end;

constructor TMSAA.Create;
begin
  inherited Create('MSAA');
end;

procedure TMSAA.MainLoop;
const
  sampleCount = 4;
var
  MSAATexture: array[0..1] of TTexture;
  MSAADepthBuffer: array[0..1] of TDepthBuffer;
  eye: NativeInt;
  destIndex: Integer;
  dstTex: ID3D11Resource;
begin
  Layer[0] := TVRLayer.Create(Session);

  // Make MSAA textures and depth buffers
  for eye := 0 to 1 do
  begin
    MSAATexture[eye] := TTexture.Create(True, Layer[0].pEyeRenderTexture[eye].SizeW,
                                              Layer[0].pEyeRenderTexture[eye].SizeH, 0, sampleCount);
    MSAADepthBuffer[eye] := TDepthBuffer.Create(DIRECTX.Device, Layer[0].pEyeRenderTexture[eye].SizeW,
                                                                Layer[0].pEyeRenderTexture[eye].SizeH, sampleCount);
  end;

  while HandleMessages do
  begin
    ActionFromInput;
    Layer[0].GetEyePoses;

    for eye := 0 to 1 do
    begin
      if DIRECTX.Key['1'] then
      begin
        // Without MSAA, for comparison
        Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye);
      end
      else begin
        // Render to higher resolution texture
        Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye, MSAATexture[eye].TexRtv, nil, 1, 1, 1, 1, 1, 0.2, 1000.0, True, MSAADepthBuffer[eye]);
        // Then resolve down into smaller buffer
        destIndex := 0;
        ovr_GetTextureSwapChainCurrentIndex(Session, Layer[0].pEyeRenderTexture[eye].TextureChain, destIndex);
        ovr_GetTextureSwapChainBufferDX(Session, Layer[0].pEyeRenderTexture[eye].TextureChain, destIndex, ID3D11Resource, dstTex);
        DIRECTX.Context.ResolveSubresource(dstTex, 0, MSAATexture[eye].Tex, 0, DXGI_FORMAT_R8G8B8A8_UNORM);
        dstTex := nil;
        Layer[0].pEyeRenderTexture[eye].Commit;
      end;
    end;

    Layer[0].PrepareLayerHeader;
    DistortAndPresent(1);
  end;

  for eye := 0 to 1 do
  begin
    MSAATexture[eye].Free;
    MSAADepthBuffer[eye].Free;
  end;
end;

procedure Main;
var
  app: TMSAA;
begin
  app := TMSAA.Create;
  try
    ExitCode := app.Run;
  finally
    app.Free;
  end;
end;

end.
