(************************************************************************************
Filename    :   Win32_RoomTiny_Main.cpp
Content     :   First-person view test application for Oculus Rift
Created     :   18th Dec 2014
Authors     :   Tom Heath
Copyright   :   Copyright 2012 Oculus, Inc. All Rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
/// A simple demo to show the extra code needed to have the player's movement
/// (from WASD and cursors) in the direction of where the player is looking (including Rift
/// orientations), not just in the direction of the player (independent of Rift orientation)
/// as exhibited in most of these samples.
unit UDemoMotionInLookDirection;

interface

procedure Main;

implementation

uses
  Winapi.Windows, System.Math.Vectors, OVR_CAPI, UDirectXAppUtil, UBasicVR;

type
  TMotionInLookDirection = class(TBasicVR)
  public
    constructor Create;
    procedure MainLoop; override;
  end;

constructor TMotionInLookDirection.Create;
begin
  inherited Create('Motion In Look Direction');
end;

procedure TMotionInLookDirection.MainLoop;
var
  totalRot: TQuaternion3D;
  fwd, right: TPoint3D;
  speed: Single;
  eye: NativeInt;
begin
  Layer[0] := TVRLayer.Create(Session);

  while HandleMessages do
  begin
    // We pass in zero, to make no positional movement
    // but keep rest of the motion intact.
    ActionFromInput(0);
    Layer[0].GetEyePoses;

    // Find the orthogonal vectors resulting from combined rift and user yaw
    totalRot := TQuaternion3D(Layer[0].EyeRenderPose[0].Orientation) * MainCam.Rot;
    fwd := Point3D(0, 0, -1) * TMatrix3D(totalRot);
    right := Point3D(1, 0, 0) * TMatrix3D(totalRot);

    // Keyboard inputs to adjust player position, using these orthogonal vectors
    speed := 0.05;
    if DIRECTX.Key['W'] or DIRECTX.Key[Chr(VK_UP)]   then MainCam.Pos := MainCam.Pos + (fwd   * +speed);
    if DIRECTX.Key['S'] or DIRECTX.Key[Chr(VK_DOWN)] then MainCam.Pos := MainCam.Pos + (fwd   * -speed);
    if DIRECTX.Key['D']                         then MainCam.Pos := MainCam.Pos + (right * +speed);
    if DIRECTX.Key['A']                         then MainCam.Pos := MainCam.Pos + (right * -speed);

    for eye := 0 to 1 do
    begin
      Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye);
    end;

    Layer[0].PrepareLayerHeader;
    DistortAndPresent(1);
  end;
end;

procedure Main;
var
  app: TMotionInLookDirection;
begin
  app := TMotionInLookDirection.Create;
  try
    ExitCode := app.Run;
  finally
    app.Free;
  end;
end;

end.