(************************************************************************************
Filename    :   Win32_RoomTiny_Main.cpp
Content     :   First-person view test application for Oculus Rift
Created     :   18th Dec 2014
Authors     :   Tom Heath
Copyright   :   Copyright 2012 Oculus, Inc. All Rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*************************************************************************************)
/// A sample to show the use of the SDK function, to recenter the pose, by pressing '1'.
/// In practise it is fairly straightforward to implement your own
/// recentering functions by intercepting Rift poses and modifying them
/// accordingly. In fact, that may be advisable as this function does one
/// specific implementation of recentering that may not be exactly
/// what you require.
unit UDemoRecenterPose;

interface

procedure Main;

implementation

uses
  Winapi.D3D11, Winapi.DxgiFormat, OVR_CAPI, OVR_CAPI_D3D, UDirectXAppUtil, UBasicVR;

type
  TRecenterPose = class(TBasicVR)
  public
    constructor Create;
    procedure MainLoop; override;
  end;

constructor TRecenterPose.Create;
begin
  inherited Create('Recenter Pose');
end;

procedure TRecenterPose.MainLoop;
var
  eye: NativeInt;
begin
  Layer[0] := TVRLayer.Create(Session);

  while HandleMessages do
  begin
    ActionFromInput;

    // Recenter the Rift by pressing '1'
    if DIRECTX.Key['1'] then
      ovr_RecenterTrackingOrigin(Session);

    Layer[0].GetEyePoses;

    for eye := 0 to 1 do
    begin
      Layer[0].RenderSceneToEyeBuffer(MainCam, RoomScene, eye);
    end;

    Layer[0].PrepareLayerHeader;
    DistortAndPresent(1);
  end;
end;

procedure Main;
var
  app: TRecenterPose;
begin
  app := TRecenterPose.Create;
  try
    ExitCode := app.Run;
  finally
    app.Free;
  end;
end;

end.