unit UYgModels;

interface

uses
  Winapi.Windows, Winapi.D3D11, System.SysUtils, System.Classes, System.Generics.Collections,
  System.Math.Vectors, Vcl.Graphics, UDirectXAppUtil,
  ygirdlib, ygtmdlib, ygmotionlib, ygswplib, ygbaselib, tgaimage, UMathHelper;

type
  TIrdType = (itAvatar, itCLT, itEnv, itFont, itItem, itMonster, itMovie, itNPC,
    itSkill, itSound, itLmt, itMap, itActive, itArtificial, itChair, itEffect,
    itFx, itNature, itPointLight, itPortal, itSky, itStructure, itTexture,
    itWater, itTile, itUI);

  TIrdManager = class
  private
  const
    IRD_PATHS: array[TIrdType] of string = ('Avatar\Avatar', 'CLT\CLT', 'Env\Env',
      'Font\Font', 'Item\Item', 'Monster\Monster', 'Movie\Movie', 'NPC\NPC',
      'Skill\Skill', 'Sound\Sound', 'Terrain\Lmt\Lmt', 'Terrain\Map',
      'Terrain\Object\Active\Active', 'Terrain\Object\Artificial\Artificial',
      'Terrain\Object\Chair\Chair', 'Terrain\Object\Effect\Effect',
      'Terrain\Object\Fx\Fx', 'Terrain\Object\Nature\Nature',
      'Terrain\Object\PointLight\PointLight', 'Terrain\Object\Portal\Portal',
      'Terrain\Object\Sky\Sky', 'Terrain\Object\Structure\Structure',
      'Terrain\Object\Texture\Texture', 'Terrain\Object\Water\Water',
      'Terrain\Tile\Tile', 'UI\UI');
  var
    FModelsPath: string;
    FCaches: TObjectDictionary<TIrdType,TIrd>;
  public
    constructor Create(const AModelsPath: string);
    destructor Destroy; override;
    function TryGetStream(AIrdType: TIrdType; const AFileName: string; AStream: TStream): Boolean;
    function LoadFile<T: TYgFile,constructor>(AIrdType: TIrdType; const AFileName: string; out AFile: T): Boolean;
  end;

  TSwapParts = TArray<Integer>;

  TSwapManager = class
  private
    FSwaps: array[Boolean] of TSwp;
  public
    constructor Create;
    destructor Destroy; override;
    function TryGetParts(AMale: Boolean; const AFileName: string; out AParts: TSwapParts): Boolean;
  end;

  PBone = ^TBone;
  TBone = record
  public
    Parent: PBone;
    Position: TPoint3D;
    Rotation: TQuaternion3D;
    Scaling: TPoint3D;
    ScalingRotation: TQuaternion3D;
    DefaultWorldMatrix: TMatrix3D;
    StaticLocalMatrix: TMatrix3D;
    WorldMatrix: TMatrix3D;
    InvertMatrix: TMatrix3D;
    SkinMatrix: TMatrix3D;
  end;

  TBoneWeight = record
  public
    BoneID: Integer; // TODO: PBoneɂďɉ
    Weight: Single;
  end;

  TYgModel = class;
  TYgModelPart = class(TModel)
  private
    FParent: TYgModel;
    FBoneWeights: array of array of TBoneWeight;
    FDefaultVertices: TArray<TVertex>;
    FSkinMatrix: TMatrix3D;
  public
    constructor Create;
    procedure Update; override;
  end;

  TYgModel = class(TRenderItem)
  private
    FModels: TObjectList<TYgModelPart>;
    FBones: TArray<TBone>;
    FMotions: TArray<TBonePoses>;
//    FBaseMotionPos: TArray<TPoint3D>;
    FFrameIndex: Single;
    //FFrameIndex: Integer;
    FFrameCount: Integer;
    FScl: Single;
    FSkinColor: UInt32;
    FBlendSkinColor: Boolean;
    function LoadTexture(IrdType: TIrdType; const FileName: string): TTexture;
    procedure LoadTMD(IrdType: TIrdType; const FileName, MotionFileName: string);
    procedure LoadFrame;
  protected
    procedure Update; override;
  public
    constructor CreateMob(const FileName: string; const APos: TPoint3D;
      const ARot: TQuaternion3D);
    constructor CreateNpc(AId: Integer; const APos: TPoint3D;
      const ARot: TQuaternion3D; const AScl: Single = 1.0);
    destructor Destroy; override;
    procedure NextFrame;
    procedure Render(const projView: TMatrix3D; R: Single; G: Single; B: Single;
      A: Single; standardUniforms: Boolean); override;
    procedure RenderInstanced(const viewProjMats: array of TMatrix3D; R: Single;
      G: Single; B: Single; A: Single; standardUniforms: Boolean); override;
  end;

  TYgModelPC = class(TYgModel)
  private
    FMale: Boolean;
  public
    constructor Create(AMale: Boolean; const APos: TPoint3D;
      const ARot: TQuaternion3D);
  end;

var
  AppPath: string;
  YgPath: string;
  YgModelsPath: string;
  IrdManager: TIrdManager;

// TODO: TIrdManager
// TODO: TYgModel.LoadPRT
// TODO: TYgModelPC.IsSwapItem(const Name: string): Boolean;
// TODO: }`eNX`Ή
// Ql: http://blog.techlab-xe.net/archives/337

implementation

function YgMatrixToMatrix(AMatrix: TTmdMatrix): TMatrix3D;
var
  row, col: Integer;
begin
  Result := TMatrix3D.Identity;
  for row := 0 to 3 do
    for col := 0 to 2 do
      Result.M[row].V[col] := AMatrix.Vectors[row][col];
end;

{ TIrdManager }

constructor TIrdManager.Create(const AModelsPath: string);
begin
  inherited Create;
  FCaches := TObjectDictionary<TIrdType,TIrd>.Create([doOwnsValues]);
  FModelsPath := AModelsPath;
end;

destructor TIrdManager.Destroy;
begin
  FCaches.Free;
  inherited;
end;

function TIrdManager.TryGetStream(AIrdType: TIrdType; const AFileName: string;
  AStream: TStream): Boolean;
var
  ird: TIrd;
  path: string;
  fs: TFileStream;
  oldpos: Int64;
begin
  path := AppPath + ExtractFileName(AFileName);
  if FileExists(path) then
  begin
    oldpos := AStream.Position;
    fs := TFileStream.Create(path, fmOpenRead or fmShareDenyWrite);
    try
      AStream.CopyFrom(fs, 0);
      AStream.Position := oldpos;
      Result := True;
    finally
      fs.Free;
    end;
    Exit;
  end;

  if not FCaches.TryGetValue(AIrdType, ird) then
  begin
    path := FModelsPath + IRD_PATHS[AIrdType] + '.IRH';
    if not FileExists(path) then
      raise Exception.Create(IRD_PATHS[AIrdType] + '.IRD not found.');

    ird := TIrd.Create;
    ird.LoadFromFile(path);
    FCaches.Add(AIrdType, ird);
  end;

  oldpos := AStream.Position;
  Result := ird.GetStream(AFileName, AStream);
  AStream.Position := oldpos;
end;

function TIrdManager.LoadFile<T>(AIrdType: TIrdType; const AFileName: string;
  out AFile: T): Boolean;
var
  ird: TIrd;
  path: string;
  ms: TMemoryStream;
begin
  Result := False;
  if FileExists(path) then
  begin
    ms := TMemoryStream.Create;
    try
      ms.LoadFromFile(path);
      ms.Position := 0;
      AFile := T.Create;
      try
        AFile.LoadFromStream(ms);
        Result := True;
      except
        AFile.Free;
      end;
    finally
      ms.Free;
    end;
    Exit;
  end;

  if not FCaches.TryGetValue(AIrdType, ird) then
  begin
    path := FModelsPath + IRD_PATHS[AIrdType] + '.IRH';
    if not FileExists(path) then
      raise Exception.Create(IRD_PATHS[AIrdType] + '.IRD not found.');

    ird := TIrd.Create;
    ird.LoadFromFile(path);
    FCaches.Add(AIrdType, ird);
  end;

  ms := TMemoryStream.Create;
  try
    if ird.GetStream(AFileName, ms) then
    begin
      ms.Position := 0;
      AFile := T.Create;
      try
        AFile.LoadFromStream(ms);
        Result := True;
      except
        AFile.Free;
      end;
    end;
  finally
    ms.Free;
  end;
end;

{ TSwapManager }

constructor TSwapManager.Create;
begin
  inherited Create;
  if not IrdManager.LoadFile(itAvatar, 'female.swp', FSwaps[False]) then
      raise Exception.Create('Can''t load female.swp');
  if not IrdManager.LoadFile(itAvatar, 'male.swp', FSwaps[True]) then
      raise Exception.Create('Can''t load male.swp');
end;

destructor TSwapManager.Destroy;
begin
  FSwaps[False].Free;
  FSwaps[True].Free;
  inherited;
end;

function TSwapManager.TryGetParts(AMale: Boolean; const AFileName: string;
  out AParts: TSwapParts): Boolean;
var
  i: Integer;
  part: TSwpSwapPart;
begin
  i := FSwaps[AMale].IndexOf(AFileName);
  if i = -1 then Exit(False);
  part := FSwaps[AMale].SwapParts[i];
  SetLength(AParts, part.PartsNum);
  for i := 0 to part.PartsNum-1 do
    AParts[i] := part.PartsData[i];
  Result := True;
end;

{ TYgModelPart }

constructor TYgModelPart.Create;
begin
  inherited Create;
end;

procedure TYgModelPart.Update;
var
  map: D3D11_MAPPED_SUBRESOURCE;
  i: NativeInt;
  p: PVertex;
  boneIdx1, boneIdx2, boneIdx3, boneIdx4: Integer;
  pos1, pos2, pos3, pos4: TPoint3D;
begin
  inherited;

  // _XV
  if Succeeded(DIRECTX.Context.Map(FVertexBuffer.Buffer, 0, D3D11_MAP_WRITE_DISCARD, 0, map)) then
  try
    p := map.pData;
    for i := 0 to High(FDefaultVertices) do
    begin
      case Length(FBoneWeights[i]) of
        1:
        begin
          boneIdx1 := FBoneWeights[i][0].BoneID;
          p^.Pos := FDefaultVertices[i].Pos * FParent.FBones[boneIdx1].SkinMatrix;
        end;
        2:
        begin
          boneIdx1 := FBoneWeights[i][0].BoneID;
          boneIdx2 := FBoneWeights[i][1].BoneID;
          pos1 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx1].SkinMatrix) * FBoneWeights[i][0].Weight;
          pos2 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx2].SkinMatrix) * FBoneWeights[i][1].Weight;
          p^.Pos := pos1 + pos2;
        end;
        3:
        begin
          boneIdx1 := FBoneWeights[i][0].BoneID;
          boneIdx2 := FBoneWeights[i][1].BoneID;
          boneIdx3 := FBoneWeights[i][2].BoneID;
          pos1 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx1].SkinMatrix) * FBoneWeights[i][0].Weight;
          pos2 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx2].SkinMatrix) * FBoneWeights[i][1].Weight;
          pos3 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx3].SkinMatrix) * FBoneWeights[i][2].Weight;
          p^.Pos := pos1 + pos2 + pos3;
        end;
        4:
        begin
          boneIdx1 := FBoneWeights[i][0].BoneID;
          boneIdx2 := FBoneWeights[i][1].BoneID;
          boneIdx3 := FBoneWeights[i][2].BoneID;
          boneIdx4 := FBoneWeights[i][3].BoneID;
          pos1 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx1].SkinMatrix) * FBoneWeights[i][0].Weight;
          pos2 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx2].SkinMatrix) * FBoneWeights[i][1].Weight;
          pos3 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx3].SkinMatrix) * FBoneWeights[i][2].Weight;
          pos4 := (FDefaultVertices[i].Pos * FParent.FBones[boneIdx4].SkinMatrix) * FBoneWeights[i][3].Weight;
          p^.Pos := pos1 + pos2 + pos3 + pos4;
        end;
      else
        raise ENotImplemented.Create('bone weights > 4');
      end;
      p^.C := FDefaultVertices[i].C;
      p^.U := FDefaultVertices[i].U;
      p^.V := FDefaultVertices[i].V;
      Inc(p);
    end;
  finally
    DIRECTX.Context.Unmap(FVertexBuffer.Buffer, 0);
  end;
end;

{ TYgModel }

constructor TYgModel.CreateMob(const FileName: string; const APos: TPoint3D;
  const ARot: TQuaternion3D);
begin
  inherited Create;

  FModels := TObjectList<TYgModelPart>.Create;
  FPos := APos;
  FRot := ARot;
  FScl := 1.0;

  LoadTMD(itMonster, FileName, ChangeFileExt(FileName, '.mlib'));
end;

constructor TYgModel.CreateNpc(AId: Integer; const APos: TPoint3D;
  const ARot: TQuaternion3D; const AScl: Single = 1.0);
var
  fname: string;
begin
  inherited Create;

  FModels := TObjectList<TYgModelPart>.Create;
  FPos := APos;
  FRot := ARot;
  FScl := AScl;

  fname := Format('cn%.4d\cn%.4d.tmd', [AId, AId]);
  LoadTMD(itNPC, fname, ChangeFileExt(fname, '.mlib'));
end;

destructor TYgModel.Destroy;
begin
  FModels.Free;
  inherited;
end;

function TYgModel.LoadTexture(IrdType: TIrdType; const FileName: string): TTexture;
var
  ms: TMemoryStream;
  ext: string;
  bmp: TBitmap;
  tga: TTGAImage;

  procedure FillTextureByBitmap(ATex: TTexture; ABitmap: TBitmap);
  var
    buf: array of DWORD;
    x, y: NativeInt;
    pix: PRGBQuad;
    src24: PRGBTriple;
    src32: PRGBQuad;
    alpha: Single;
  begin
    SetLength(buf, ABitmap.Width * ABitmap.Height);
    if ABitmap.PixelFormat = pf32bit then
    begin
      for y := 0 to ABitmap.Height-1 do
      begin
        src32 := ABitmap.ScanLine[y];
        pix := @buf[ABitmap.Width * y];
        for x := 0 to ABitmap.Width-1 do
        begin
          if not FBlendSkinColor then
          begin
            pix^.rgbBlue := src32^.rgbRed;
            pix^.rgbGreen := src32^.rgbGreen;
            pix^.rgbRed := src32^.rgbBlue;
            pix^.rgbReserved := src32^.rgbReserved;
          end
          else begin
            alpha := src32^.rgbReserved / 255;
            pix^.rgbBlue := TRGBQuad(FSkinColor).rgbBlue + Trunc((src32^.rgbRed - TRGBQuad(FSkinColor).rgbBlue) * alpha);
            pix^.rgbGreen := TRGBQuad(FSkinColor).rgbGreen + Trunc((src32^.rgbGreen - TRGBQuad(FSkinColor).rgbGreen) * alpha);
            pix^.rgbRed := TRGBQuad(FSkinColor).rgbRed + Trunc((src32^.rgbBlue - TRGBQuad(FSkinColor).rgbRed) * alpha);
            pix^.rgbReserved := $FF;
          end;
          Inc(src32);
          Inc(pix);
        end;
      end;
    end
    else begin
      ABitmap.PixelFormat := pf24bit;
      for y := 0 to ABitmap.Height-1 do
      begin
        src24 := ABitmap.ScanLine[y];
        pix := @buf[ABitmap.Width * y];
        for x := 0 to ABitmap.Width-1 do
        begin
          pix^.rgbBlue := src24^.rgbtRed;
          pix^.rgbGreen := src24^.rgbtGreen;
          pix^.rgbRed := src24^.rgbtBlue;
          pix^.rgbReserved := $FF;
          Inc(src24);
          Inc(pix);
        end;
      end;
    end;
    ATex.FillTexture(@buf[0]);
  end;

begin
  ms := TMemoryStream.Create;
  try
    if not IrdManager.TryGetStream(IrdType, FileName, ms) then
      raise Exception.Create('Texture not found.');

    ext := LowerCase(ExtractFileExt(FileName));
    if ext = '.bmp' then
    begin
      bmp := TBitmap.Create;
      try
        bmp.LoadFromStream(ms);
        Result := TTexture.Create(bmp.Width, bmp.Height, False);
        try
          FillTextureByBitmap(Result, bmp);
        except
          FreeAndNil(Result);
          raise;
        end;
      finally
        bmp.Free;
      end;
    end
    else if ext = '.tga' then
    begin
      tga := TTGAImage.Create;
      try
        tga.LoadFromStream(ms);
        Result := TTexture.Create(tga.Width, tga.Height, False);
        try
          FillTextureByBitmap(Result, tga.Bitmap);
        except
          FreeAndNil(Result);
          raise;
        end;
      finally
        tga.Free;
      end;
    end
    else
      raise Exception.Create('Texture file type is not supported.');
  finally
    ms.Free;
  end;
end;

procedure TYgModel.LoadTMD(IrdType: TIrdType; const FileName, MotionFileName: string);
var
  ms: TMemoryStream;
  tmd: TTmd;
  iSkin, iMeshRow: Integer;
  materialIndex: Integer;
  texpath: string;
  t: TTriangleSet;
  i, j: NativeInt;
  part: TYgModelPart;
  skin: TTmdObjectContainer;
  stdmesh: TTmdStandardMesh;
  indices: TArray<SmallInt>;
//  xobj: TTmdXFormObject;
  bone: TTmdObjectTrackItem;
  mlib: TMlib;
  motionIndex: Integer;
//  pos1, pos2: TPoint3D;
begin
  tmd := TTmd.Create;
  try
    ms := TMemoryStream.Create;
    try
      if not IrdManager.TryGetStream(IrdType, FileName, ms) then
        raise Exception.Create('TMD not found.');
      ms.Position := 0;
      tmd.LoadFromStream(ms);
    finally
      ms.Free;
    end;

    for iSkin := 0 to tmd.Main.ObjectBlock.ObjectTrack.SkinCount-1 do
    begin
      skin := tmd.Main.ObjectBlock.ObjectTrack.Skins[iSkin];
      for iMeshRow := 0 to skin.MeshRowCount-1 do
      begin
        part := TYgModelPart.Create;
        try
          part.FParent := Self;
          materialIndex := skin.MeshRows[iMeshRow].StandardMesh.MaterialNumber.Number;

          if not (tmd.Main.ObjectBlock.MaterialTrack.Items[materialIndex] is TTmdMaterial) then
            raise Exception.Create('Materials block not found.');
          if TTmdMaterial(tmd.Main.ObjectBlock.MaterialTrack.Items[materialIndex]).HasTexture = 0 then
          begin
            if (string(TTmdMaterial(tmd.Main.ObjectBlock.MaterialTrack.Items[materialIndex]).Name) = 'face') and (Self is TYgModelPC) then
            begin
              //part.Fill := TMaterial.Create(CreateSolidTexture(32, 32));
              if TYgModelPC(Self).FMale then
                texpath := 'face\male\male_face_M0010.tga'
              else
                texpath := 'face\female\female_face_F0010.tga';
              part.Fill := TMaterial.Create(LoadTexture(IrdType, texpath));
            end
            else
              raise Exception.Create('No texture.');
          end
          else begin
            texpath := ExtractFilePath(FileName) + string(TTmdMaterial(tmd.Main.ObjectBlock.MaterialTrack.Items[materialIndex]).TextureName);
            part.Fill := TMaterial.Create(LoadTexture(IrdType, texpath));
          end;

          // _ǉ
          t := Default(TTriangleSet);
          stdmesh := skin.MeshRows[iMeshRow].StandardMesh;
          SetLength(part.FDefaultVertices, Length(stdmesh.VertexList.Vectors));
          for i := 0 to High(stdmesh.VertexList.Vectors) do
          begin
            part.FDefaultVertices[i].Pos.V[0] := stdmesh.VertexList.Vectors[i][0];
            part.FDefaultVertices[i].Pos.V[1] := stdmesh.VertexList.Vectors[i][1];
            part.FDefaultVertices[i].Pos.V[2] := stdmesh.VertexList.Vectors[i][2];
            part.FDefaultVertices[i].C := $FFFFFFFF;
            part.FDefaultVertices[i].U := stdmesh.TextureVertexList.Vectors[i][0];
            part.FDefaultVertices[i].V := stdmesh.TextureVertexList.Vectors[i][1];
          end;
          t.Vertices := part.FDefaultVertices;
          SetLength(indices, Length(stdmesh.FaceList.Vectors)*3);
          for i := 0 to High(stdmesh.FaceList.Vectors) do
          begin
            indices[i*3+0] := stdmesh.FaceList.Vectors[i][0];
            indices[i*3+1] := stdmesh.FaceList.Vectors[i][1];
            indices[i*3+2] := stdmesh.FaceList.Vectors[i][2];
          end;
          t.Indices := indices;

  //        for i := 0 to tmd.Main.ObjectBlock.ObjectTrack.Items.Count-1 do
  //        begin
  //          if tmd.Main.ObjectBlock.ObjectTrack.Items[i].ID <> TMD_BONE then Continue;
  //
  //          bone := TTmdObjectTrackItem(tmd.Main.ObjectBlock.ObjectTrack.Items[i]);
  //          t.AddSolidColorBox(
  //            bone.WorldMatrix.Vectors[3][0] + 0.01,
  //            bone.WorldMatrix.Vectors[3][1] - 0.01,
  //            bone.WorldMatrix.Vectors[3][2] + 0.01,
  //            bone.WorldMatrix.Vectors[3][0] - 0.01,
  //            bone.WorldMatrix.Vectors[3][1] + 0.01,
  //            bone.WorldMatrix.Vectors[3][2] - 0.01,
  //            $FFFF0000
  //          );
  //        end;

          part.Init(t);

          // {[EFCgێ
          if stdmesh.SoftSkinTrack <> nil then
          begin
            SetLength(part.FBoneWeights, Length(stdmesh.SoftSkinTrack.Weights));
            for i := 0 to High(stdmesh.SoftSkinTrack.Weights) do
            begin
              SetLength(part.FBoneWeights[i], Length(stdmesh.SoftSkinTrack.Weights[i]));
              for j := 0 to High(stdmesh.SoftSkinTrack.Weights[i]) do
              begin
                part.FBoneWeights[i][j].BoneID := stdmesh.SoftSkinTrack.BoneIDs[stdmesh.SoftSkinTrack.Weights[i][j].BoneIndex];
                part.FBoneWeights[i][j].Weight := stdmesh.SoftSkinTrack.Weights[i][j].Weight;
              end;
            end;
          end
          else if stdmesh.RigidSkinTrack <> nil then
          begin
            SetLength(part.FBoneWeights, Length(stdmesh.RigidSkinTrack.BoneIndex));
            for i := 0 to High(stdmesh.RigidSkinTrack.BoneIndex) do
            begin
              SetLength(part.FBoneWeights[i], 1);
              part.FBoneWeights[i][0].BoneID := stdmesh.RigidSkinTrack.BoneIDs[stdmesh.RigidSkinTrack.BoneIndex[i]];
              part.FBoneWeights[i][0].Weight := 1.0;
            end;
          end
          else
            raise Exception.Create('No weights');

          part.FSkinMatrix := YgMatrixToMatrix(skin.WorldMatrix);

          FModels.Add(part);
        except
          part.Free;
        end;
      end; // for iMeshRow
    end; // for iSkin

    // {[s
{
    SetLength(FBones, tmd.Main.XFormBlock.Items.Count);
    for i := 0 to tmd.Main.XFormBlock.Items.Count-1 do
    begin
      xobj := TTmdXFormObject(tmd.Main.XFormBlock.Items[i]);
      bone := TTmdObjectTrackItem(tmd.Main.ObjectBlock.ObjectTrack.Items[i]);

      if xobj.XFormObject_ParentNum = nil then
        FBones[i].Parent := nil
      else
        FBones[i].Parent := @FBones[xobj.XFormObject_ParentNum.Number];

      if xobj.PositionList <> nil then
        raise ENotImplemented.Create('')
      else
        FBones[i].Position := TPoint3D.Zero;

      if xobj.RotateList <> nil then
        raise ENotImplemented.Create('')
      else
        FBones[i].Rotation := TQuaternion3D.Identity;

//        if xobj.ScaleList <> nil then
//          raise ENotImplemented.Create('')
//        else
        FBones[i].Scaling := TPoint3D.Create(1, 1, 1);

//        if xobj.ScaleAxisList <> nil then
//          raise ENotImplemented.Create('')
//        else
        FBones[i].ScalingRotation := TQuaternion3D.Identity;

      FBones[i].DefaultWorldMatrix := YgMatrixToMatrix(bone.WorldMatrix);
      FBones[i].StaticLocalMatrix := YgMatrixToMatrix(bone.StaticLocalMatrix);
      FBones[i].InvertMatrix := FBones[i].DefaultWorldMatrix.Inverse;
    end;
}
    SetLength(FBones, tmd.Main.ObjectBlock.ObjectTrack.Items.Count);
    for i := 0 to tmd.Main.ObjectBlock.ObjectTrack.Items.Count-1 do
    begin
      bone := TTmdObjectTrackItem(tmd.Main.ObjectBlock.ObjectTrack.Items[i]);
      FBones[i].DefaultWorldMatrix := YgMatrixToMatrix(bone.WorldMatrix);
      FBones[i].StaticLocalMatrix := YgMatrixToMatrix(bone.StaticLocalMatrix);
      FBones[i].InvertMatrix := FBones[i].DefaultWorldMatrix.Inverse;

      FBones[i].Position := TPoint3D.Zero;
      FBones[i].Rotation := TQuaternion3D.Identity;
      FBones[i].Scaling := TPoint3D.Create(1, 1, 1);
      FBones[i].ScalingRotation := TQuaternion3D.Identity;
    end;
  finally
    tmd.Free;
  end;

  mlib := TMlib.Create;
  try
    ms := TMemoryStream.Create;
    try
      if not IrdManager.TryGetStream(IrdType, MotionFileName, ms) then
        raise Exception.Create('Motion not found.');
      ms.Position := 0;
      mlib.LoadFromStream(ms);

      for i := 0 to High(mlib.ParentBoneIndices) do
      begin
        if mlib.ParentBoneIndices[i] < 0 then
          FBones[i].Parent := nil
        else
          FBones[i].Parent := @FBones[mlib.ParentBoneIndices[i]];
      end;

      motionIndex := 29;
      if (motionIndex < 0) or (motionIndex >= mlib.Motions.Count) then
        motionIndex := 2;
      if (motionIndex < 0) or (motionIndex >= mlib.Motions.Count) then
        motionIndex := Random(mlib.Motions.Count);
      FMotions := Copy(mlib.Motions[motionIndex].FBones);
//        FBaseMotionPos := Copy(mlib.Motions[2].FUnknown4);
      FFrameCount := Length(mlib.Motions[motionIndex].FUnknown4);
      FFrameIndex := 0;
      LoadFrame;
    finally
      ms.Free;
    end;
  finally
    mlib.Free;
  end;
end;

procedure TYgModel.LoadFrame;
var
  i, j: NativeInt;
  rate: Single;
begin
  {
  for i := 0 to High(FMotions) do
  begin
    for j := 0 to High(FMotions[i].Positions) do
    begin
      if FMotions[i].Positions[j].Frame = FFrameIndex then
      begin
        FBones[i].Position := FMotions[i].Positions[0].Point;
        Break;
      end
      else if FMotions[i].Positions[j].Frame > FFrameIndex then
      begin
        if FFrameIndex = 0 then
          FBones[i].Position := TPoint3D.Zero;
        Break;
      end;
    end;

    for j := 0 to High(FMotions[i].Rotations) do
    begin
      if FMotions[i].Rotations[j].Frame = FFrameIndex then
      begin
        FBones[i].Rotation := FMotions[i].Rotations[j].Quaternion;
        Break;
      end
      else if FMotions[i].Rotations[j].Frame > FFrameIndex then
      begin
        if FFrameIndex = 0 then
          FBones[i].Rotation := TQuaternion3D.Identity;
        Break;
      end;
    end;
  end;
  }
  for i := 0 to High(FMotions) do
  begin
    if Length(FMotions[i].Positions) = 0 then
      FBones[i].Position := TPoint3D.Zero
    else if Length(FMotions[i].Positions) = 1 then
      FBones[i].Position := FMotions[i].Positions[0].Point
    else begin
      for j := 0 to High(FMotions[i].Positions)-1 do
      begin
        if (FFrameIndex >= FMotions[i].Positions[j].Frame) and (FFrameIndex <= FMotions[i].Positions[j+1].Frame) then
        begin
          //FBones[i].Position := FMotions[i].Positions[j].Point;
          rate := (FFrameIndex-FMotions[i].Positions[j].Frame) / (FMotions[i].Positions[j+1].Frame-FMotions[i].Positions[j].Frame);
          FBones[i].Position := FMotions[i].Positions[j].Point.Lerp(FMotions[i].Positions[j+1].Point, rate);
          Break;
        end
        else if j = High(FMotions[i].Positions)-1 then
        begin
          FBones[i].Position := FMotions[i].Positions[j+1].Point;
          Break;
        end;
      end;
    end;

    if Length(FMotions[i].Rotations) = 0 then
      FBones[i].Rotation := TQuaternion3D.Identity
    else if Length(FMotions[i].Rotations) = 1 then
      FBones[i].Rotation := FMotions[i].Rotations[0].Quaternion
    else begin
      for j := 0 to High(FMotions[i].Rotations)-1 do
      begin
        if (FFrameIndex >= FMotions[i].Rotations[j].Frame) and (FFrameIndex <= FMotions[i].Rotations[j+1].Frame) then
        begin
          //FBones[i].Rotation := FMotions[i].Rotations[j].Quaternion;
          rate := (FFrameIndex-FMotions[i].Rotations[j].Frame) / (FMotions[i].Rotations[j+1].Frame-FMotions[i].Rotations[j].Frame);
          FBones[i].Rotation := TQuaternion3D.Slerp(FMotions[i].Rotations[j].Quaternion, FMotions[i].Rotations[j+1].Quaternion, rate);
          Break;
        end
        else if j = High(FMotions[i].Rotations)-1 then
        begin
          FBones[i].Rotation := FMotions[i].Rotations[j+1].Quaternion;
          Break;
        end;
      end;
    end;

    if Length(FMotions[i].Scalings) = 0 then
      FBones[i].Scaling := TPoint3D.Create(1,1,1)
    else if Length(FMotions[i].Scalings) = 1 then
      FBones[i].Scaling := FMotions[i].Scalings[0].Point
    else begin
      for j := 0 to High(FMotions[i].Scalings)-1 do
      begin
        if (FFrameIndex >= FMotions[i].Scalings[j].Frame) and (FFrameIndex <= FMotions[i].Scalings[j+1].Frame) then
        begin
          rate := (FFrameIndex-FMotions[i].Scalings[j].Frame) / (FMotions[i].Scalings[j+1].Frame-FMotions[i].Scalings[j].Frame);
          FBones[i].Scaling := FMotions[i].Scalings[j].Point.Lerp(FMotions[i].Scalings[j+1].Point, rate);
          Break;
        end
        else if j = High(FMotions[i].Scalings)-1 then
        begin
          FBones[i].Scaling := FMotions[i].Scalings[j+1].Point;
          Break;
        end;
      end;
    end;

    if Length(FMotions[i].ScalingRotations) = 0 then
      FBones[i].ScalingRotation := TQuaternion3D.Identity
    else if Length(FMotions[i].ScalingRotations) = 1 then
      FBones[i].ScalingRotation := FMotions[i].ScalingRotations[0].Quaternion
    else begin
      for j := 0 to High(FMotions[i].ScalingRotations)-1 do
      begin
        if (FFrameIndex >= FMotions[i].ScalingRotations[j].Frame) and (FFrameIndex <= FMotions[i].ScalingRotations[j+1].Frame) then
        begin
          rate := (FFrameIndex-FMotions[i].ScalingRotations[j].Frame) / (FMotions[i].ScalingRotations[j+1].Frame-FMotions[i].ScalingRotations[j].Frame);
          FBones[i].ScalingRotation := TQuaternion3D.Slerp(FMotions[i].ScalingRotations[j].Quaternion, FMotions[i].ScalingRotations[j+1].Quaternion, rate);
          Break;
        end
        else if j = High(FMotions[i].ScalingRotations)-1 then
        begin
          FBones[i].ScalingRotation := FMotions[i].ScalingRotations[j+1].Quaternion;
          Break;
        end;
      end;
    end;
  end;
end;

procedure TYgModel.NextFrame;
begin
  FFrameIndex := FFrameIndex + 0.33333;
  if FFrameIndex >= FFrameCount then
    FFrameIndex := FFrameIndex - FFrameCount;
  {
  FFrameIndex := (FFrameIndex + 1) mod FFrameCount;
  }
  LoadFrame;
end;

procedure TYgModel.Update;
var
  iBone, iModel: NativeInt;

  function GetParentWorldMatrix(const ABone: TBone): TMatrix3D;
  var
//    pos1, pos2: TPoint3D;
    localMat: TMatrix3D;
  begin
//    pos1 := ABone.Position;
//    if ABone.Parent <> nil then
//    begin
//      pos2 := ABone.Parent^.Position;
//      pos1 := pos1 - pos2;
//    end;
//
////    localMat := TMatrix3D(ABone.Rotation) * ABone.DefaultWorldMatrix;
////    localMat := TMatrix3D(ABone.Rotation) * ABone.DefaultWorldMatrix;
//    localMat := TMatrix3D(ABone.Rotation) * TMatrix3D.CreateTranslation(pos1);
//
//    if ABone.Parent <> nil then
//      Result := localMat * GetParentWorldMatrix(ABone.Parent^)
//    else
//      Result := localMat * ABone.DefaultWorldMatrix;
{
    localMat := TMatrix3D(ABone.Rotation) * TMatrix3D.CreateTranslation(ABone.Position) * ABone.StaticLocalMatrix;
    if ABone.Parent <> nil then
      Result := localMat * GetParentWorldMatrix(ABone.Parent^)
    else
      Result := localMat;
}

    localMat := TMatrix3D(ABone.ScalingRotation).Inverse * TMatrix3D.CreateScaling(ABone.Scaling) * TMatrix3D(ABone.ScalingRotation) * TMatrix3D(ABone.Rotation) * TMatrix3D.CreateTranslation(ABone.Position);
    if ABone.Parent <> nil then
      Result := localMat * GetParentWorldMatrix(ABone.Parent^) //(ABone.Parent^.DefaultWorldMatrix)
    else
      //Result := TMatrix3D(ABone.Rotation) * TMatrix3D.CreateTranslation(FBaseMotionPos[Trunc(FFrameIndex)]) //localMat;
      Result := localMat;
  end;

begin
  if FModels.Count = 1 then
  begin
    // O[osXV
    for iBone := Low(FBones) to High(FBones) do
    begin
      FBones[iBone].WorldMatrix := GetParentWorldMatrix(FBones[iBone]);
      FBones[iBone].SkinMatrix := FBones[iBone].InvertMatrix * FBones[iBone].WorldMatrix * TMatrix3D.CreateScaling(Point3D(1,1,-1));
    end;

    // _XV
    for iModel := 0 to FModels.Count-1 do
    begin
      FModels[iModel].Update;
    end;
  end
  else begin
    // O[osXV
    for iBone := Low(FBones) to High(FBones) do
    begin
      FBones[iBone].WorldMatrix := GetParentWorldMatrix(FBones[iBone]);
    end;

    for iModel := 0 to FModels.Count-1 do
    begin
      // XLsXV
      for iBone := Low(FBones) to High(FBones) do
      begin
        FBones[iBone].SkinMatrix := FModels[iModel].FSkinMatrix * FBones[iBone].InvertMatrix * FBones[iBone].WorldMatrix * TMatrix3D.CreateScaling(Point3D(1,1,-1));
      end;

      // _XV
      FModels[iModel].Update;
    end;
  end;
end;

procedure TYgModel.Render(const projView: TMatrix3D; R, G, B, A: Single;
  standardUniforms: Boolean);
var
  i: NativeInt;
  mat: TMatrix3D;
begin
  mat := TMatrix3D.CreateScaling(Point3D(FScl,FScl,FScl)) * TMatrix3D(FRot) * TMatrix3D.CreateTranslation(FPos) * projView;
  for i := 0 to FModels.Count-1 do
    FModels[i].Render(mat, R, G, B, A, standardUniforms);
end;

procedure TYgModel.RenderInstanced(const viewProjMats: array of TMatrix3D; R, G,
  B, A: Single; standardUniforms: Boolean);
var
  i: NativeInt;
  mats: array of TMatrix3D;
begin
  SetLength(mats, Length(viewProjMats));
  for i := Low(viewProjMats) to High(viewProjMats) do
    mats[i] := TMatrix3D(FRot) * TMatrix3D.CreateTranslation(FPos) * viewProjMats[i];

  for i := 0 to FModels.Count-1 do
    FModels[i].RenderInstanced(mats, R, G, B, A, standardUniforms);
end;

{ TYgModelPC }

constructor TYgModelPC.Create(AMale: Boolean; const APos: TPoint3D;
  const ARot: TQuaternion3D);
begin
  inherited Create;

  FModels := TObjectList<TYgModelPart>.Create;
  FPos := APos;
  FRot := ARot;
  FScl := 1.0;

  FSkinColor := $FFC1E4FE; //$C1E4FE, $90C9FF, $4C7BB8
  FBlendSkinColor := True;

  FMale := AMale;
  if FMale then
    LoadTMD(itAvatar, 'male.TMD', 'motion\male\male.mlib')
  else
    LoadTMD(itAvatar, 'female.TMD', 'motion\female\female.mlib');
end;

initialization
  AppPath := ExtractFilePath(ParamStr(0));
  YgPath := 'D:\Games\Yogurting\';
  YgModelsPath := YgPath + 'Models\';
  ygtmdlib.DisableReverseZ := True;
  IrdManager := TIrdManager.Create(YgModelsPath);
finalization
  IrdManager.Free;
end.